"""
https://gist.github.com/ibeltagy/23c1d71f2494837e91dcecd613c8b5cb

Original by @ibeltagy
Modified by @soldni
"""

from typing import List, Optional

from shadow_scholar.cli import safe_import

with safe_import():
    import torch

    # from accelerate import init_empty_weights
    from transformers import AutoModelForCausalLM, AutoTokenizer


# def replace_8bit_linear(
#     model, threshold=6.0, module_to_not_convert="lm_head"
# ):
#     for name, module in model.named_children():
#         if len(list(module.children())) > 0:
#             replace_8bit_linear(module, threshold, module_to_not_convert)

#         if (
#             isinstance(module, torch.nn.Linear)
#             and name != module_to_not_convert
#         ):
#             with init_empty_weights():
#                 model._modules[name] = bnb.nn.Linear8bitLt(
#                     module.in_features,
#                     module.out_features,
#                     module.bias is not None,
#                     has_fp16_weights=False,
#                     threshold=threshold,
#                 )
#     return model


def assert_all_approx_close(a, b, rtol, atol, count):
    idx = torch.isclose(a.float(), b.float(), rtol, atol)
    sumval = (idx == 0).sum().item()
    if sumval > count:
        print(f"Too many values not close: assert {sumval} < {count}")
        try:
            torch.testing.assert_allclose(a, b, rtol, atol)
        except Exception as e:
            print(e)


def get_memory_footprint(model, return_buffers=True):
    """
    Get the memory footprint of a model. This will return the memory footprint
    of the current model in bytes. Useful to benchmark the memory footprint of
    the current model and design some tests. Solution inspired from the
    PyTorch discussions:
    https://discuss.pytorch.org/t/gpu-memory-that-model-uses/56822/2

    Arguments:
        return_buffers (`bool`, *optional*, defaults to `True`): Whether to
            return the size of the buffer tensors in the computation of the
            memory footprint. Buffers are tensors that do not require gradients
            and not registered as parameters. E.g. mean and std in batch norm
            layers.
    """
    mem = sum(
        [
            param.nelement() * param.element_size()
            for param in model.parameters()
        ]
    )
    if return_buffers:
        mem_bufs = sum(
            [buf.nelement() * buf.element_size() for buf in model.buffers()]
        )
        mem = mem + mem_bufs
    return mem


def replace_linear_with_int8(
    model: "torch.nn.Module", do_not_convert: Optional[List[str]] = None
):
    do_not_convert = do_not_convert or ["lm_head"]

    names = [name for name, _ in model.named_modules()]

    for name in names:
        if any(name.startswith(e) for e in do_not_convert):
            continue

        module = model.get_submodule(name)

        if isinstance(module, torch.nn.Linear):
            new_module = QuantizedLinearInt8(layer=module)
            parent_name, module_name = name.rsplit(".", 1)
            parent = model.get_submodule(parent_name)
            delattr(parent, module_name)
            setattr(parent, module_name, new_module)

        torch.cuda.empty_cache()


class QuantizedLinearInt8(torch.nn.Module):
    """
    A simple but effictive implmenetion of Int8 quantization for linear layers.
    The weights are quantized and stored as Int8, which saves ~50% of the gpu
    memory. During the forwared pass, the weights are de-quantized back to fp16
    to do multiplication.

    Pros:
        - saves ~50% of the gpu memory
        - accurate quantization because only the weights are quantized, and
            the weights don't suffer from the "outliers" issue mentioned in
            the LLM.int8 paper; only the activations do.
        - high precision results beacuse the multiplication is done in fp16
        - much faster than LLM.int8
    Cons:
        - a bit slower because of the added computation of dequantization
            in each forward pass. In practice, the slowdown is not large
            because in the generation application, gpu utilization is not
            very high.
    """

    def __init__(self, layer, weight_bit_width: int = 8):
        super().__init__()

        self.bias = torch.nn.Parameter(
            layer.bias.clone(), requires_grad=layer.bias.requires_grad
        )

        self.weight_scale = torch.nn.Parameter(
            (
                layer.weight.abs().max(dim=-1).values
                / ((2 ** (weight_bit_width - 1)) - 1)
            ).half(),
        )

        self.weight = torch.nn.Parameter(
            torch.round(
                layer.weight.float() / self.weight_scale[:, None]
            ).char(),
            requires_grad=False,
        )

    def forward(self, x):
        # TODO: the following two lines can be combined in a single more
        # efficient command. One way to do so is torch.einsum (as in the
        # commented code below), but the order of operations in einsum is
        # arbitrary, so it can multiply `x * self.weight` before
        # `self.weight_scale`, which results in higher loss of precision.

        # The dequantization is two steps:
        # 1) `self.weight.half()`: which casts the weights back to fp16
        # 2) `* self.weight_scale[:, None]` multiply the weights by the
        #       quantization weight scales
        # Profiling shows that (1) is the main reason for the slowdown, and
        # that the overhead from (2) is very small

        weight = self.weight.half() * self.weight_scale[:, None]
        return torch.nn.functional.linear(x, weight, self.bias)

        # if x.dim() == 2:
        #     result = torch.einsum(
        #         "sh, dh, d -> sd", x, self.weight.half(), self.weight_scale
        #     )
        # elif x.dim() == 3:
        #     result = torch.einsum(
        #         "bsh, dh, d -> bsd", x, self.weight.half(), self.weight_scale
        #     )
        # else:
        #     # Not supported
        #     print(x)
        #     print(x.shape)
        #     assert False

        # if self.bias is not None:
        #     result += self.bias
        # return result


def convert_model_to_int8_on_gpu(model, device):
    """
    Quantize a model to int8 and move it to GPU using a simple method.
    """
    if "cuda" not in device:
        raise ValueError(
            f"Target device should be a gpu. Device {device} is not supported"
        )

    model.half()

    memory_before_quantization = get_memory_footprint(model)  # without lm_head

    # replace `Linear` with `QuantizedLinearInt8`
    replace_linear_with_int8(model)

    model.to(device=device)
    memory_after_quantization = get_memory_footprint(model)  # without lm_head

    saving = round(
        100 * memory_after_quantization / memory_before_quantization
    )
    memory_before_quantization = round(
        memory_before_quantization / 2**30, 2
    )  # rounding for printing
    memory_after_quantization = round(
        memory_after_quantization / 2**30, 2
    )  # rounding for printing

    print(
        f"Quantization memory - before: {memory_before_quantization} GB, "
        f"after: {memory_after_quantization} GB ({saving}% of the size before)"
    )
    return model


def main():
    model_name = "facebook/opt-350m"
    tokenizer = AutoTokenizer.from_pretrained(model_name)
    input_ids = tokenizer.encode(
        "The sky is", return_tensors="pt"
    ).cuda()  # pyright: ignore

    model_nonint8 = AutoModelForCausalLM.from_pretrained(model_name)
    model_nonint8 = model_nonint8.half().cuda()

    model_int8 = AutoModelForCausalLM.from_pretrained(model_name)
    model_int8 = convert_model_to_int8_on_gpu(model_int8, device="cuda")

    with torch.inference_mode():
        logits_nonint8 = model_nonint8(input_ids=input_ids).logits
        logits_int8 = model_int8(input_ids=input_ids).logits
        print(torch.mean(abs(logits_int8 - logits_nonint8)).detach().item())
        assert_all_approx_close(
            logits_int8, logits_nonint8, rtol=0.01, atol=3.0e-2, count=30
        )

        generated_nonint8 = model_nonint8.generate(input_ids)
        generated_int8 = model_int8.generate(input_ids)
        print("=============")
        print(tokenizer.decode(generated_nonint8[0]))
        print("=============")
        print(tokenizer.decode(generated_int8[0]))
        print("=============")

        print(
            "Memory of non-int8 model: "
            f"{get_memory_footprint(model_nonint8) / 10**9}G"
        )
        print(
            f"Memory of int8 model: "
            f"{get_memory_footprint(model_int8) / 10**9}G"
        )


if __name__ == "__main__":
    main()



from abc import ABC, abstractmethod
from typing import List, Union

from shadow_scholar.cli import safe_import

from .datasets import Document
from .reg_utils import CallableRegistry

with safe_import():
    import spacy


slicer_registry: CallableRegistry["BaseSlicer"] = CallableRegistry()


class BaseSlicer(ABC):
    """Abstract base class for slicers.
    A slicer is a class that can slice a document into a list of slices;
    each slice is a list of spans and/or boxes, alongside the text of
    the slice."""

    @abstractmethod
    def slice_one(self, doc: Document) -> List[Document]:
        ...

    def slice(
        self, seq_or_doc: Union[Document, List[Document]]
    ) -> List[Document]:
        if isinstance(seq_or_doc, Document):
            return self.slice_one(seq_or_doc)
        else:
            return [doc for doc in seq_or_doc for doc in self.slice_one(doc)]


@slicer_registry.add("sent")
class SentenceSlicer(BaseSlicer):
    """A slicer that slices a document into a single sentence."""

    stride: int
    nlp: "spacy.language.Language"

    def __init__(
        self,
        spacy_model: str = "en_core_web_sm",
        stride: int = 0,
    ):
        self.stride = stride
        self.nlp = spacy.load(spacy_model)

    def slice_text(self, doc: Document) -> List[Document]:
        docs = []
        for sent in self.nlp(doc.text).sents:
            if self.stride > 0:
                text = sent.doc[
                    max(sent.start - self.stride, 0) : min(
                        sent.end + self.stride, len(sent.doc)
                    )
                ].text
            else:
                text = sent.text
            docs.append(doc.new(text=str(text)))
        return docs

    def slice_one(self, doc: Document) -> List[Document]:
        if hasattr(doc, "span_groups"):
            raise NotImplementedError(
                "SentenceSlicer does not support slicing documents with "
                "span groups."
            )
        else:
            return self.slice_text(doc)


@slicer_registry.add("block")
class BlockSlicer(SentenceSlicer):
    def __init__(
        self,
        length: int = 300,
        stride: Union[int, float] = 1.0,
        spacy_model: str = "en_core_web_sm",
    ):
        self.length = length

        if isinstance(stride, float):
            assert 0 <= stride <= 1, "overlap must be between 0 and 1"
            stride = int(stride * length)

        super().__init__(spacy_model=spacy_model, stride=stride)

    def slice_text(self, doc: Document) -> List[Document]:
        docs = []
        parsed_doc = self.nlp(doc.text)
        for i in range(0, len(parsed_doc), self.stride):
            span = parsed_doc[i : i + self.length]
            docs.append(doc.new(text=span.text))
        return docs



import argparse
import json
import os
import re
import urllib.parse
from dataclasses import dataclass, field
from time import time
from typing import (
    Dict,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    Union,
    cast,
)

import requests

from .library import S2_API_KEY, Paper, get_s2_metadata
from .logging import get_logger

S2_SEARCH_ENDPOINT = "https://api.semanticscholar.org/graph/v1/paper/search/"
GOOGLE_SEARCH_API_KEY = os.environ.get("GOOGLE_SEARCH_API_KEY", "")


class BaseSearchEndpoint:
    def __init__(self, endpoint: str, api_key: str):
        self.endpoint = endpoint
        self.api_key = api_key
        self.logger = get_logger(__name__, disable_cloudwatch="yes")

    def to_json(self) -> Dict[str, str]:
        return {"endpoint": self.endpoint, "name": self.__class__.__name__}

    def encode(self, text: str) -> str:
        text = re.sub(r"^\-\s*", "", text)
        text = re.sub(r"[\"\']+", "", text)
        text = re.sub(r"(^\s+|\s+$)", "", text)
        return text

    def __call__(self, query: str, max_results: int = 1000) -> List[Paper]:
        raise NotImplementedError()


class S2SearchEndpoint(BaseSearchEndpoint):
    def __init__(
        self,
        endpoint: str = S2_SEARCH_ENDPOINT,
        api_key: str = S2_API_KEY,
    ):
        super().__init__(endpoint=endpoint, api_key=api_key)

    def encode(self, text: str) -> str:
        text = super().encode(text)
        return text.replace(" ", "+")

    def __call__(self, query: str, max_results: int = 1000) -> List[Paper]:
        start = time()
        query = self.encode(query)
        url = f"{self.endpoint}?query={query}"
        headers = {"x-api-key": self.api_key}
        response = requests.get(url, headers=headers).json()
        end = time()
        self.logger.info(
            {
                "action": "s2_search",
                "query": query,
                "elapsed": end - start,
                "response": response,
            }
        )

        papers = [
            Paper(
                p["paperId"], metadata={"score": 1.0 / i, "title": p["title"]}
            )
            for i, p in enumerate(response.get("data", []))
        ]
        return papers[:max_results]


class GoogleSearchEndpoint(BaseSearchEndpoint):
    def __init__(
        self,
        endpoint: str = "https://www.googleapis.com/customsearch/v1",
        api_key: str = GOOGLE_SEARCH_API_KEY,
        cx: str = "602714345f3a24773",
    ):
        super().__init__(endpoint=endpoint, api_key=api_key)
        self.cx = cx

    def encode(self, text: str) -> str:
        text = super().encode(text)
        return urllib.parse.quote(text)

    def __call__(self, query: str, max_results: int = 1000) -> List[Paper]:
        start = time()
        query = self.encode(query)
        url = (
            f"https://www.googleapis.com/customsearch/v1/siterestrict?"
            f"&key={self.api_key}"
            f"&cx={self.cx}"
            f"&q={query}"
        )
        response = requests.get(url).json()
        end = time()

        self.logger.info(
            {
                "action": "google_search",
                "query": query,
                "elapsed": end - start,
                "response": response,
            }
        )
        papers = [
            Paper.from_url(url=r["link"], score=1.0 / i, title=r["title"])
            for i, r in enumerate(
                cast(list, response.get("items", [])), start=1
            )
        ]
        filtered = [p for p in papers if p is not None][:max_results]
        return filtered


class Query(NamedTuple):
    user_query: Union[str, None]
    system_query: Union[str, None]

    def __bool__(self):
        return self.text != ""

    @property
    def text(self) -> str:
        return self.user_query or self.system_query or ""

    def to_json(self) -> Dict[str, Union[str, None]]:
        return {
            "user_query": self.user_query,
            "system_query": self.system_query,
        }

    @classmethod
    def from_json(cls, data: Dict[str, Union[str, None]]):
        return cls(**data)


@dataclass
class Queries:
    queries: Dict[str, Query] = field(default_factory=dict)
    used: Set[Tuple[str, Optional[int]]] = field(default_factory=set)

    def add(self, queries: List[Tuple[Union[str, None], Union[str, None]]]):
        for query in queries:
            query = Query(*query)
            if not query:
                continue
            self.queries[query.text] = query

    def use(self, num_results: Optional[int] = None) -> Iterable[str]:
        for query in self.queries:
            key = (query, num_results)
            if key not in self.used:
                self.used.add(key)
                yield query

    def list(self) -> List[Tuple[Union[str, None], Union[str, None]]]:
        return list(self.queries.values())

    def to_json(
        self,
    ) -> Dict[str, List[Tuple[Union[str, None], Union[str, None]]]]:
        return {"queries": self.list()}

    @classmethod
    def from_json(
        cls, data: Dict[str, List[Tuple[Union[str, None], Union[str, None]]]]
    ):
        queries = cls()
        queries.add(data["queries"])
        return queries


if __name__ == "__main__":
    ap = argparse.ArgumentParser()
    ap.add_argument(
        "-e",
        "--endpoint",
        type=str,
        default="google",
        choices=["google", "s2"],
    )
    ap.add_argument("-q", "--query", type=str, required=True)
    ap.add_argument("-m", "--max_results", type=int, default=1000)
    args = ap.parse_args()

    endpoint: BaseSearchEndpoint
    if args.endpoint == "google":
        endpoint = GoogleSearchEndpoint()
    elif args.endpoint == "s2":
        endpoint = S2SearchEndpoint()
    else:
        raise ValueError(f"Unknown endpoint {args.endpoint}")

    papers = endpoint(args.query, max_results=args.max_results)
    get_s2_metadata(papers)
    print(json.dumps([p.to_json() for p in papers], indent=2))
    print(
        f"Found {sum(1 for p in papers if not p.missing):,} papers"
        f" with metadata out of {len(papers):,} total"
    )



import itertools
from typing import Dict, Iterable, List, NamedTuple, Optional, Sequence

from shadow_scholar.cli import safe_import

with safe_import():
    from mmda.types.annotation import BoxGroup, SpanGroup  # pyright: ignore
    from mmda.types.box import Box  # pyright: ignore
    from mmda.types.document import Document  # pyright: ignore
    from mmda.types.names import Tokens  # pyright: ignore


class BoxKey(NamedTuple):
    top: float
    page: int


def box_group_from_span_group(
    span_group: "SpanGroup",
    doc: Optional["Document"] = None,
    merge_if_on_same_row: bool = True,
    position_rounding: int = 2,
) -> "BoxGroup":
    doc = span_group.doc or doc
    if doc is None:
        raise ValueError("span_group must have a doc, or doc must be provided")

    boxes_it: Iterable[Box] = (
        span.box  # pyright: ignore
        for token in doc.find_overlapping(span_group, Tokens)
        for span in token.spans
    )

    if merge_if_on_same_row:
        boxes_it = merge_box_groups_on_same_row(
            boxes=boxes_it, position_rounding=position_rounding
        )

    return BoxGroup(boxes=list(boxes_it), type=span_group.type)


def merge_box_groups_on_same_row(
    box_groups: Optional[Iterable["BoxGroup"]] = None,
    boxes: Optional[Iterable["Box"]] = None,
    position_rounding: int = 2,
) -> Sequence["Box"]:
    if box_groups is not None and boxes is not None:
        raise ValueError("either box_groups or boxes must be provided")

    to_merge: Dict[BoxKey, List[Box]] = {}

    # combine both boxes for box_groups (if provided) as well as `boxes`
    # (if provided) into a single iterable of boxes.
    boxes_it = itertools.chain(
        (box for box_group in (box_groups or []) for box in box_group.boxes),
        (boxes or []),
    )

    # to decide which box to merge, we key them of their
    for box in boxes_it:
        to_merge.setdefault(
            BoxKey(top=round(box.t, position_rounding), page=box.page), []
        ).append(box)

    merged_boxes: List[Box] = []
    for box_key, boxes in to_merge.items():
        merged_boxes.append(
            Box(
                l=(left := min(box.l for box in boxes)),
                t=(top := min(box.t for box in boxes)),
                w=(max(box.l + box.w for box in boxes) - left),
                h=(max(box.t + box.h for box in boxes) - top),
                page=box_key.page,
            )
        )

    return merged_boxes
